from spynal.spectra.spectra import spectrum, spectrogram, power_spectrum, \
                                   power_spectrogram, phase_spectrogram, \
                                   itpc, intertrial_phase_clustering, burst_analysis, \
                                   plot_spectrum, plot_spectrogram
from spynal.spectra.wavelet import wavelet_spectrum, wavelet_spectrogram, compute_wavelets, \
                                   wavelet_bandwidth, wavelet_edge_extent
from spynal.spectra.multitaper import multitaper_spectrum, multitaper_spectrogram, compute_tapers
from spynal.spectra.bandfilter import bandfilter_spectrum, bandfilter_spectrogram, bandfilter, \
                                      set_filter_params
from spynal.spectra.preprocess import cut_trials, realign_data, realign_data_on_event, \
                                      remove_dc, remove_evoked
from spynal.spectra.postprocess import one_over_f_norm, pool_freq_bands, pool_time_epochs
from spynal.spectra.utils import next_power_of_2, get_freq_sampling, \
                                 one_sided_to_two_sided, simulate_oscillation, \
                                 complex_to_spec_type, power, magnitude, phase, real, imag


# -*- coding: utf-8 -*-
""" Postprocessing functions for LFP/EEG/continuous data and spectral analysis """
import numpy as np

try:
    import xarray as xr
    HAS_XARRAY = True
except ImportError:
    HAS_XARRAY = False

from spynal.utils import set_random_seed, iarange
from spynal.spectra.helpers import _str_to_pool_func


def pool_freq_bands(data, bands, axis=None, freqs=None, func='mean'):
    """
    Pool (eg average) spectral data within each of a given set of frequency bands

    Parameters
    ----------
    data : ndarray or xarray DataArray, shape=(...,n_freqs,...)
        Raw data to pool within frequency bands. Any arbitary shape.

    bands : array-like, shape=(n_bands,2) or dict {str : array-like, shape=(2,)}
        Frequency bands to pool data within. Input either as a list of [low-cut, high-cut]
        values or as a dict, with keys being frequency band names and their associated
        values being the corresponding [low-cut, high-cut] pair.
        Band edges are inclusive.

    axis : int
        Data axis corresponding to frequency.
        Only needed if `data` is not an xarray DataArray with dimension named 'freq'/'frequency'.

    freqs : array-like, shape=(n_freqs,)
        Frequency sampling in `data`. Only needed if `data` is not an xarray DataArray.

    func : str or callable, default: 'mean' (mean across all frequencies in band)
        Function to use to pool values within each frequency band, given either as a
        string specifier (options: 'mean' or 'sum') or a custom function that takes as input
        an ndarray and returns an ndarray with its first axis reduced to length 1.

    Returns
    -------
    data : ndarray or xarray DataArray, shape=(...,n_freqbands,...)
        Data with values pooled within each of given frequency bands
    """
    # Convert list of frequency band ranges to {'name':freq_range} dict
    if not isinstance(bands,dict):
        bands = {'band_'+str(i_band):frange for i_band,frange in enumerate(bands)}

    # Convert frequency bands into 1-d list of bin edges
    bins = []
    for value in bands.values(): bins.extend(value)

    func_ = _str_to_pool_func(func)

    # Figure out data dimensionality and standardize so frequency axis = 0 (1st axis)
    if HAS_XARRAY and isinstance(data,xr.DataArray):
        dims = np.asarray(data.dims)
        # Find frequency dimension if not given explicitly
        if axis is None:  axis = ((dims == 'freq') | (dims == 'frequency')).nonzero()[0][0]
        freq_dim = dims[axis]   # Name of frequency dim

        if freqs is None: freqs = data.coords[freq_dim].values

        # Permute array dims so freq is 1st dim
        if axis != 0:
            temp_dims = np.concatenate(([dims[axis]], dims[dims != freq_dim]))
            data = data.transpose(*temp_dims)
        else:
            temp_dims = dims

        # Initialize new DataArray with freq dim = freq bands, indexed by band names
        coords = {dim : data.coords[dim].values for dim in data.coords}
        coords[freq_dim] = list(bands.keys())
        data_shape = (len(bands), *data.shape[1:])
        band_data = xr.DataArray(np.empty(data_shape, dtype=data.dtype),
                                 dims=temp_dims, coords=coords, attrs=data.attrs)

    else:
        assert axis is not None, \
        ValueError("For ndarray data, must give frequency axis in <axis>")
        assert freqs is not None, \
        ValueError("For ndarray data, must give frequency sampling vector in <freqs>")

        if axis != 0: data = data.swapaxes(0,axis)

        data_shape= (len(bands), *data.shape[1:])
        band_data = np.empty(data_shape, dtype=data.dtype)

    # Pool data over each frequency band
    for i_band,(_,frange) in enumerate(bands.items()):
        fbool = (freqs >= frange[0]) & (freqs <= frange[1])
        band_data[i_band,...] = func_(data[fbool,...])

    # Permute back to original data dimension order
    if axis != 0:
        if HAS_XARRAY and isinstance(data,xr.DataArray):
            band_data = band_data.transpose(*dims)
        else:
            band_data = band_data.swapaxes(axis,0)

    return band_data


def pool_time_epochs(data, epochs, axis=None, timepts=None, func='mean'):
    """
    Pool (eg average) spectral data within each of a given set of time epochs

    Parameters
    ----------
    data : ndarray or xarray DataArray, shape=(...,n_timepts,...)
        Raw data to pool within time epochs. Any arbitary shape.

    epochs : array-like, shape=(n_epochs,2) or dict {str : array-like, shape=(2,)}
        Time epochs to pool data within. Input either as a list of [start,end] times
        or as a dict, with keys being time epoch names and their associated
        values being the corresponding [start,end] pair.
        Epoch edges are inclusive.

    axis : int
        Data axis corresponding to time.
        Only needed if `data` is not an xarray DataArray with dimension named 'time'.

    timepts : array-like, shape(n_timepts,)
        Time sampling in `data`.
        Only needed if `data` is not an xarray DataArray with dimension named 'time'.

    func : str or callable, default: 'mean' (mean across all frequencies in band)
        Function to use to pool values within each time epoch, given either as a
        string specifier (options: 'mean' or 'sum') or a custom function that takes as input
        an ndarray and returns an ndarray with its first axis reduced to length 1.

    Returns
    -------
    data : ndarray or xarray DataArray, shape=(...,n_time_epochs,...)
        Data with values pooled within each of given time epochs
    """
    # Convert list of time epoch ranges to {'name':time_range} dict
    if not isinstance(epochs,dict):
        epochs = {'epochs_'+str(i_epoch):trange for i_epoch,trange in enumerate(epochs)}

    func_ = _str_to_pool_func(func)

    # Figure out data dimensionality and standardize so time axis = 0 (1st axis)
    if HAS_XARRAY and isinstance(data,xr.DataArray):
        dims = np.asarray(data.dims)
        if timepts is None: timepts = data.coords['time'].values
        # Find 'time' dimension if not given explicitly
        if axis is None:  axis = (dims == 'time').nonzero()[0][0]
        # Permute array dims so time is 1st dim
        if axis != 0:
            temp_dims = np.concatenate(([dims[axis]], dims[dims != 'time']))
            data = data.transpose(*temp_dims)
        else:
            temp_dims = dims

        # Initialize new DataArray with time dim = time epochs, indexed by epoch names
        coords = {dim : data.coords[dim].values for dim in data.coords}
        coords['time'] = list(epochs.keys())
        data_shape= (len(epochs), *data.shape[1:])
        epoch_data = xr.DataArray(np.empty(data_shape, dtype=data.dtype),
                                  dims=temp_dims, coords=coords, attrs=data.attrs)

    else:
        assert axis is not None, \
        ValueError("For ndarray data, must give time axis in <axis>")
        assert timepts is not None, \
        ValueError("For ndarray data, must give time sampling vector in <timepts>")

        if axis != 0: data = data.swapaxes(0,axis)

        data_shape= (len(epochs), *data.shape[1:])
        epoch_data = np.empty(data_shape, dtype=data.dtype)

    # Pool data over each time epoch
    for i_epoch,(_,trange) in enumerate(epochs.items()):
        tbool = (timepts >= trange[0]) & (timepts <= trange[1])
        epoch_data[i_epoch,...] = func_(data[tbool,...])

    # Permute back to original data dimension order
    if axis != 0:
        if HAS_XARRAY and isinstance(data,xr.DataArray):
            epoch_data = epoch_data.transpose(*dims)
        else:
            epoch_data = epoch_data.swapaxes(axis,0)

    return epoch_data


def one_over_f_norm(data, axis=None, freqs=None, exponent=1.0):
    """
    Normalize to correct for ~ 1/frequency**alpha baseline distribution of power
    by multiplying by frequency, raised to a given exponent

    Parameters
    ----------
    data : ndarray or xarray DataArray, shape=(...,n_freqs,...)
        Raw data to pool within frequency bands. Any arbitary shape.

    axis : int
        Data axis corresponding to frequency.
        Needed if `data` is not an xarray DataArray with dimension named 'freq'/'frequency'.

    freqs : array-like, shape=(n_freqs,)
        Frequency sampling in `data`. Needed if `data` is not an xarray DataArray.

    exponent : float, default: 1 (correct for 1/f, w/o exponent)
        Exponent ('alpha') to raise freqs to for normalization.

    Returns
    -------
    data : ndarray or xarray DataArray, shape=(...,n_freqs,...)
        1/f normalized data. Same shape as input.
    """
    if HAS_XARRAY and isinstance(data,xr.DataArray):
        dims = np.asarray(data.dims)
        # Find frequency dimension if not given explicitly
        if axis is None:  axis = ((dims == 'freq') | (dims == 'frequency')).nonzero()[0][0]
        freq_dim = dims[axis]   # Name of frequency dim
        if freqs is None: freqs = data.coords[freq_dim].values

    assert axis is not None, \
        ValueError("Frequency axis must be given in <axis> (or input xarray data)")
    assert freqs is not None, \
        ValueError("Frequency sampling vector must be given in <freqs> (or input xarray data)")

    # Ensure that freqs will broadcast against data
    freqs = np.asarray(freqs)
    if data.ndim != freqs.ndim:
        slicer          = [np.newaxis]*data.ndim    # Create (data.ndim,) list of np.newaxis
        slicer[axis]    = slice(None)               # Set <axis> element to slice as if set=':'
        freqs           = freqs[tuple(slicer)]      # Expand freqs to dimensionality of data

    return data * freqs**exponent




# -*- coding: utf-8 -*-
""" Helper functions for `randstats` module for randomization statistics """
import numpy as np

from spynal.utils import one_sample_tstat, two_sample_tstat, one_way_fstat, two_way_fstat, \
                         correlation, rank_correlation

#==============================================================================
# Utility functions
#==============================================================================

#==============================================================================
# Private helper functions
#==============================================================================
def _tail_to_compare(tail):
    """ Convert string specifier to callable function implementing it """
    # If input value is already a callable function, just return it
    if callable(tail): return tail

    assert isinstance(tail,str), \
        TypeError("Unsupported type '%s' for <tail>. Use string or function" % type(tail))

    tail = tail.lower()

    # 2-tailed test: hypothesis ~ stat_obs ~= statShuf
    if tail == 'both':
        return lambda stat_obs,stat_resmp: np.abs(stat_resmp) >= np.abs(stat_obs)

    # 1-tailed rightward test: hypothesis ~ stat_obs > statShuf
    elif tail == 'right':
        return lambda stat_obs,stat_resmp: stat_resmp >= stat_obs

    # 1-tailed leftward test: hypothesis ~ stat_obs < statShuf
    elif tail == 'left':
        return lambda stat_obs,stat_resmp: stat_resmp <= stat_obs

    else:
        ValueError("Unsupported value '%s' for <tail>. Use 'both', 'right', or 'left'" % tail)


def _str_to_one_sample_stat(stat, axis):
    """ Convert string specifier to function to compute 1-sample statistic """
    if callable(stat): return stat

    assert isinstance(stat,str), \
        TypeError("Unsupported type '%s' for <stat>. Use string or function" % type(stat))

    stat = stat.lower()
    if stat in ['t','tstat','t1']:      return lambda data: one_sample_tstat(data, axis=axis)
    elif stat == 'mean':                return lambda data: data.mean(axis=axis, keepdims=True)
    else:
        raise ValueError('Unsupported option ''%s'' given for <stat>' % stat)


def _str_to_assoc_stat(stat, axis):
    """ Convert string specifier to function to compute paired-sample association statistic """
    if callable(stat): return stat

    assert isinstance(stat,str), \
        TypeError("Unsupported type '%s' for <stat>. Use string or function" % type(stat))

    stat = stat.lower()
    if stat in ['r','pearson','pearsonr']:
        return lambda data1,data2: correlation(data1, data2, axis=axis)
    elif stat in ['r','pearson','pearsonr']:
        return lambda data1,data2: rank_correlation(data1, data2, axis=axis)
    else:
        raise ValueError('Unsupported option ''%s'' given for <stat>' % stat)


def _str_to_two_sample_stat(stat, axis):
    """ Convert string specifier to function to compute 2-sample statistic """
    if callable(stat): return stat

    assert isinstance(stat,str), \
        TypeError("Unsupported type '%s' for <stat>. Use string or function" % type(stat))

    stat = stat.lower()
    if stat in ['t','tstat','t1']:
        return lambda data1,data2: two_sample_tstat(data1, data2, axis=axis)
    elif stat in ['meandiff','mean']:
        return lambda data1,data2: (data1.mean(axis=axis, keepdims=True) -
                                    data2.mean(axis=axis, keepdims=True))
    else:
        raise ValueError('Unsupported option ''%s'' given for <stat>' % stat)


def _str_to_one_way_stat(stat, axis):
    """ Convert string specifier to function to compute 1-way multi-sample statistic """
    if callable(stat): return stat

    assert isinstance(stat,str), \
        TypeError("Unsupported type '%s' for <stat>. Use string or function" % type(stat))

    stat = stat.lower()
    if stat in ['f','fstat','f1']:
        return lambda data, labels: one_way_fstat(data, labels, axis=axis)
    else:
        raise ValueError('Unsupported option ''%s'' given for <stat>' % stat)


def _str_to_two_way_stat(stat, axis):
    """ Convert string specifier to function to compute 2-way multi-sample statistic """
    if callable(stat): return stat

    assert isinstance(stat,str), \
        TypeError("Unsupported type '%s' for <stat>. Use string or function" % type(stat))

    stat = stat.lower()
    if stat in ['f','fstat','f2']:
        return lambda data, labels: two_way_fstat(data, labels, axis=axis)
    else:
        raise ValueError('Unsupported option ''%s'' given for <stat>' % stat)


def _paired_sample_data_checks(data1, data2):
    """ Check data format requirements for paired-sample data """

    assert np.array_equal(data1.shape, data2.shape), \
        ValueError("data1 and data2 must have same shape for paired-sample tests. \
                    Use two-sample tests to compare non-paired data with different n's.")


def _two_sample_data_checks(data1, data2, axis):
    """ Check data format requirements for two-sample data """

    assert (data1.ndim == data2.ndim), \
        "data1 and data2 must have same shape except for observation/trial axis (<axis>)"

    if data1.ndim > 1:
        assert np.array_equal([data1.shape[ax] for ax in range(data1.ndim) if ax != axis],
                              [data2.shape[ax] for ax in range(data2.ndim) if ax != axis]), \
            "data1 and data2 must have same shape except for observation/trial axis (<axis>)"



"""
Suite of tests to assess "face validity" of randomization statistic functions in randstats.py
Usually used to test new or majorly updated functions to ensure they perform as expected.

Includes tests that parametrically estimate statistics as a function of difference in distribution
means, assays of bias, etc. to establish methods produce expected pattern of results.

Plots results and runs assertions that basic expected results are reproduced

Function list
-------------
- test_randstats :          Contains tests of randomization statistic computation functions
- stat_test_battery :       Runs standard battery of tests of randomization stat functions

- test_confints :           Contains tests of confidence interval computation functions
- confint_test_battery :    Runs standard battery of tests of confidence interval functions
"""
import os
import time
from math import ceil
import numpy as np
import matplotlib.pyplot as plt

from spynal.tests.data_fixtures import simulate_dataset
from spynal.utils import set_random_seed
from spynal.randstats.randstats import one_sample_test, paired_sample_test, paired_sample_association_test, \
                                       two_sample_test, one_way_test, two_way_test, \
                                       one_sample_confints, paired_sample_confints, two_sample_confints


# =============================================================================
# Tests for functions computing statistics/statistical tests
# =============================================================================
def test_randstats(stat, method, test='gain', test_values=None, term=0, distribution='normal',
                   alpha=0.05, n_reps=100, seed=None, do_tests=True,
                   do_plots=False, plot_dir=None, **kwargs):
    """
    Basic testing for randomization statistic computation functions

    Generates synthetic data, computes statistics, p values, and significance using given method,
    and compares computed to expected values.

    For test failures, raises an error or warning (depending on value of `do_tests`).
    Optionally plots summary of test results.

    Parameters
    ----------
    stat : str
        Type of statistical test to evaluate: 
        'one_sample'|'paired_sample'|'paired_sample_assoc'| 'two_sample'|'one_way'|'two_way'

    method : str
        Resampling paradigm to use for test: 'permutation' | 'bootstrap'

    test : str, default: 'gain'
        Type of test to run. Options:
        
        - 'gain' : Tests multiple values for btwn-cond response difference (gain).
            Checks for monotonically increasing stat/decreasing p value.
        - 'spread' : Tests multiple values for distribution spread (SD).
            Checks for monotonically decreasing stat/increasing p value.
        - 'n' : Tests multiple values of number of trials (n).
            Checks that stat doesn't vary, but p value decreases, with n.
        - 'bias' : Tests multiple n values with 0 btwn-cond difference.
            Checks that stat doesn't vary and p value remains ~ 0 (unbiased).

    test_values : array-like, shape=(n_values,), dtype=str
        List of values to test. Interpretation and defaults are test-specific:
        
        - 'gain' :      Btwn-condition response differences (gains). Default: [1,2,5,10,20]
        - 'spread' :    Gaussian SDs for each response distribution. Default: [1,2,5,10,20]
        - 'n'/'bias' :  Trial numbers. Default: [25,50,100,200,400,800]

    term : int, default: 0
        Which model term to modify for testing 2-way stats (unused for other stats).
        0,1 = main effects, 2 = interaction. 

    distribution : str, default: 'normal'
        Name of distribution to simulate data from. Options: 'normal' | 'poisson'

    n_reps : int, default: 100
        Number of independent repetitions of tests to run

    alpha : float, default: 0.05
        Significance criterion "alpha"

    do_tests : bool, default: True
        Set=True to evaluate test results against expected values and raise an error if they fail

    do_plots : bool, default: False
        Set=True to plot test results

    plot_dir : str, default: None (don't save to file)
        Full-path directory to save plots to. Set=None to not save plots.

    seed : int, default: 1 (reproducible random numbers)
        Random generator seed for repeatable results. Set=None for fully random numbers.

    **kwargs :
        All other keyword args passed as-is to `simulate_dataset` or statistic computation
        function, as appropriate

    Returns
    -------
    means : dict, {str : ndarray. shape=(n_values,)}
        Mean results (across independent test runs) of variables output from randomization tests
        for each tested value. Each key/value pair corressponds to a computed statistic variable:
        
        - 'signif' :    Binary signficance decision (at criterion <alpha>)
        - 'p' :         p values
        - 'log_p' :     p values negative log-transformed -log10(p) to increase with effect size
        - 'stat_obs' :  Observed evaluatation statistic values
        - 'stat_resmp' :Mean resampled statistic values (across all resamples)

    sds : : dict, {str : ndarray. shape=(n_values,)}
        Standard deviation of results (across test runs) of variables output from rand tests for
        each tested value. Same fields as means.

    passed : bool
        True if all tests produce expected values; otherwise False.
    """
    # Note: Set random seed once here, not for every random data generation loop below
    if seed is not None: set_random_seed(seed)

    test = test.lower()
    method = method.lower()

    # Set defaults for tested values and set up data generator function depending on <test>
    # Note: Only set random seed once above, don't reset in data generator function calls
    if stat == 'two_way':       n_conds = 4         # Simulate 2-way design w/ 4 conds (2x2)
    elif stat == 'one_sample':  n_conds = 1
    else:                       n_conds = 2
    if stat == 'two_way':
        # Set gains for effect on 1st main effect, 2nd main effect, and interaction effect
        if term == 0:   gain_pattern = np.asarray([0,0,1,1])
        elif term == 1: gain_pattern = np.asarray([1,0,1,0])
        elif term == 2: gain_pattern = np.asarray([0,1,1,0])
    else:
        gain_pattern = 1
        
    sim_args = dict(gain=5.0*gain_pattern, offset=0.0, spreads=10.0, n_conds=n_conds, n=100,
                    distribution=distribution, correlation=0, seed=None)
    # Override defaults with any simulation-related params passed to function
    for arg in sim_args:
        if arg in kwargs: sim_args[arg] = kwargs.pop(arg)

    if test == 'gain':
        test_values = [1,2,5,10,20] if test_values is None else test_values
        del sim_args['gain']                        # Delete preset arg so uses arg to lambda below
        gen_data = lambda gain: simulate_dataset(**sim_args,gain=gain*gain_pattern)

    elif test in ['spread','spreads','sd']:
        test_values = [1,2,5,10,20] if test_values is None else test_values
        del sim_args['spreads']                     # Delete preset arg so uses arg to lambda below
        gen_data = lambda spreads: simulate_dataset(**sim_args,spreads=spreads)

    elif test in ['n','n_trials','bias']:
        test_values = [25,50,100,200,400,800] if test_values is None else test_values
        if test == 'bias': sim_args.update(offset=0, gain=0)    # Set gain,offset=0 for bias test
        del sim_args['n']                           # Delete preset arg so uses arg to lambda below
        # Set model correlation != 0 for paired_sample_assoc stat to see effects of changing n
        if test in ['n','n_trials'] and (stat == 'paired_sample_assoc'): sim_args['correlation'] = 0.2
        gen_data = lambda n_trials: simulate_dataset(**sim_args,n=n_trials)

    elif test == 'n_conds':
        test_values = [2,4,8] if test_values is None else test_values
        del sim_args['n_conds']                     # Delete preset arg so uses arg to lambda below
        gen_data = lambda n_conds: simulate_dataset(**sim_args,n_conds=n_conds)

    elif test == 'correlation':
        test_values = [0,0.25,0.5,0.75,1] if test_values is None else test_values
        del sim_args['correlation']                 # Delete preset arg so uses arg to lambda below
        gen_data = lambda correlation: simulate_dataset(**sim_args,correlation=correlation)

    else:
        raise ValueError("Unsupported value '%s' set for <test>" % test)

    # Set up function for computing randomization stats
    stat_func = _str_to_stat_func(stat)
    kwargs.update(return_stats=True)
    if 'n_resamples' not in kwargs: kwargs['n_resamples'] = 100

    n_terms = 3 if stat == 'two_way' else 1
    results = dict(signif = np.empty((n_terms,len(test_values),n_reps),dtype=bool),
                   p = np.empty((n_terms,len(test_values),n_reps)),
                   log_p = np.empty((n_terms,len(test_values),n_reps)),
                   stat_obs = np.empty((n_terms,len(test_values),n_reps)),
                   stat_resmp = np.empty((n_terms,len(test_values),n_reps)))

    resmp_axis = -1 if stat == 'two_way' else 0

    for i_value,test_value in enumerate(test_values):
        for i_rep in range(n_reps):
            # Generate simulated data with current test value
            data,labels = gen_data(test_value)

            # One-sample tests
            if stat == 'one_sample':
                p, stat_obs, stat_resmp = stat_func(data, method=method, **kwargs)

            # 1-way and 2-way ANOVA-like multi-level factorial tests
            elif stat in ['one_way','two_way']:
                # For 2-way tests, reorg labels from (n,) vector of values in set {0-3}
                # to (n,3) array where 1st 2 col's are 2 orthogonal factors, 3rd col is interaction
                if stat == 'two_way':
                    labels = np.stack((labels >= 2, (labels == 1) | (labels == 3), labels), axis=1)
                p, stat_obs, stat_resmp = stat_func(data, labels, method=method, **kwargs)

            # Paired-sample and two-sample tests
            else:
                p, stat_obs, stat_resmp = stat_func(data[labels==1], data[labels==0],
                                                    method=method, **kwargs)


            # Determine which values are significant (p < alpha criterion)
            results['signif'][:,i_value,i_rep]      = p < alpha
            results['p'][:,i_value,i_rep]           = p
            # Negative log-transform p values so increasing = "better" and less compressed
            results['log_p'][:,i_value,i_rep]       = -np.log10(p)
            results['stat_obs'][:,i_value,i_rep]    = stat_obs
            # Compute mean resampled stat value across all resamples
            results['stat_resmp'][:,i_value,i_rep]  = stat_resmp.mean(axis=resmp_axis)


    # Compute mean and std dev across different reps of simulation
    means   = {variable : values.mean(axis=-1) for variable,values in results.items()}
    sds     = {variable : values.std(axis=-1,ddof=0) for variable,values in results.items()}
    variables = ['signif','log_p','stat_obs','stat_resmp']

    if do_plots:
        plt.figure()
        for i_vbl,variable in enumerate(variables):
            for i_term in range(n_terms):
                mean = means[variable][i_term,:]
                sd = sds[variable][i_term,:]

                sp = i_term*len(variables) + i_vbl + 1
                plt.subplot(n_terms, len(variables), sp)
                plt.grid(axis='both',color=[0.75,0.75,0.75],linestyle=':')
                plt.errorbar(test_values, mean, sd, marker='o')
                if i_term == n_terms-1:
                    plt.xlabel('n' if test == 'bias' else test)
                    plt.ylabel('-log10(p)' if variable == 'log_p' else variable)
                if i_term == 0: plt.title(variable)
                if stat == 'two_way': plt.ylabel("Term %d" % i_term)

        if plot_dir is not None:
            filename = os.path.join(plot_dir,'stat-summary-%s-%s-%s' % (stat,method,test))
            if stat == 'two_way': filename += '-term%d' % term
            plt.savefig(filename)

    # Determine if test actually produced the expected values
    # 'gain' : Test if p values decrease and stat increases monotonically with between-group gain
    if test == 'gain':
        if stat != 'paired_sample_assoc':
            evals = [((np.diff(means['signif'][term,:]) >= 0).all(),
                        "Significance does not increase monotonically with btwn-cond mean diff"),
                    ((np.diff(means['log_p'][term,:]) >= 0).all(),
                        "p values do not decrease monotonically with btwn-cond mean diff"),
                    ((np.diff(means['stat_obs'][term,:]) > 0).all(),
                        "Statistic does not increase monotonically with btwn-cond mean diff"),
                    (means['stat_resmp'][term,:].ptp() <= sds['stat_resmp'][term,:].max(),
                        "Resampled stat has larger than expected range with btwn-cond mean diff")]
        else:
            evals = [] # todo should we test for no change here?
            
    # 'spread' : Test if p val's increase and stat decreases monotonically with within-group spread
    elif test in ['spread','spreads','sd']:
        if stat != 'paired_sample_assoc':
            evals = [((np.diff(means['signif'][term,:]) > 0).all(),
                        "Signif does not decrease monotonically with within-cond spread increase"),
                    ((np.diff(means['log_p'][term,:]) >= 0).all(),
                        "p values do not increase monotonically with within-cond spread increase"),
                    ((np.diff(means['stat_obs'][term,:]) < 0).all(),
                        "Statistic does not decrease monotonically with within-cond spread increase"),
                    (means['stat_resmp'][term,:].ptp() <= sds['stat_resmp'].max(),
                        "Resampled stat has > than expected range with within-cond spread increase")]
        else:
            evals = [] # todo should we test for no change here?

    # 'n' : Test if p values decrease, but statistic is ~ same for all values of n (unbiased by n)
    elif test in ['n','n_trials']:
        evals = [((np.diff(means['signif'][term,:]) >= 0).all(),
                    "Significance does not increase monotonically with n"),
                 ((np.diff(means['log_p'][term,:]) >= 0).all(),
                    "p values do not decrease monotonically with n"),
                 ((np.diff(means['stat_obs'][term,:]) >= 0).all(),
                    "Statistic does not decrease monotonically with n"),
                 (means['stat_resmp'][term,:].ptp() <= sds['stat_resmp'].max(),
                    "Resampled stat has > expected range across n's (likely biased by n)")]
        if stat == 'paired_sample_assoc': del evals[2]
        
    # 'bias': Test that statistic is not > 0 and p value ~ alpha if gain = 0, for varying n
    elif test == 'bias':
        evals = [(((np.abs(means['signif'][term,:].mean()) - alpha) < sds['signif'][term,:]).all(),
                    "Signif different from expected pct (%.1f) when no mean diff between conds"
                    % alpha),
                 (((np.abs(means['p'][term,:].mean()) - 0.5) < sds['p'][term,:]).all(),
                    "p values are different from expected (0.5) when no mean diff between conds"),
                 ((np.abs(means['stat_obs'][term,:]) < sds['stat_obs'][term,:]).all(),
                    "Statistic is above 0 when no mean diff between conds")]

    # 'correlation': Test that correlation stat monotonically increases with correction,
    # that other statistics are ~ same for all values of correlation
    elif test == 'correlation':
        if stat == 'paired_sample_assoc':
            evals = [((np.diff(means['signif'][term,:]) >= 0).all(),
                        "Significance does not increase monotonically with correlation"),
                    ((np.diff(means['log_p'][term,:]) >= 0).all(),
                        "p values do not decrease monotonically with correlation"),
                    ((np.diff(means['stat_obs'][term,:]) > 0).all(),
                        "Statistic does not increase monotonically with correlation"),
                    (means['stat_resmp'][term,:].ptp() <= sds['stat_resmp'][term,:].max(),
                        "Resampled stat has larger than expected range with correlation")]                     
        else:
            evals = [] # todo Should we test for no change otherwise
            
        
    passed = True
    for cond,message in evals:
        if not cond:    passed = False

        # Raise an error for test fails if do_tests is True
        if do_tests:    assert cond, AssertionError(message)
        # Just issue a warning for test fails if do_tests is False
        elif not cond:  print("Warning: " + message)

    return means, sds, passed


def stat_test_battery(stats=('one_sample','paired_sample','paired_sample_assoc',
                             'two_sample','one_way','two_way'),
                      methods=('permutation','bootstrap'),
                      tests=('gain','spread','n','bias','correlation'),
                      do_tests=True, **kwargs):
    """
    Run a battery of given tests on given randomization statistic computation methods

    Parameters
    ----------
    stats : array-like of str, default: ('one_sample','paired_sample','paired_sample_assoc',
        'two_sample','one_way','two_way')
        List of statistical tests to evaluate.

    methods : array-like of str, default: ('permutation','bootstrap') (all supported methods)
        List of resampling paradigms to run.
                
    tests : array-like of str, default: ('gain','spread','n','bias','correlation')
        List of tests to run.
                
    do_tests : bool, default: True
        Set=True to evaluate test results against expected values and raise an error if they fail

    **kwargs :
        Any other kwargs passed directly to test_randstats()
    """
    if isinstance(stats,str): stats = [stats]
    if isinstance(methods,str): methods = [methods]
    if isinstance(tests,str): tests = [tests]

    for stat in stats:
        for test in tests:
            for method in methods:
                # TEMP Bootstrap version of 1-way/2-way tests not coded up yet...
                if (stat in ['one_way','two_way']) and (method != 'permutation'): continue
                print("Running %s test on %s %s" % (test,stat,method))

                # Run separate tests for each term of 2-way stats (2 main effects and interaction)
                if stat == 'two_way':
                    passed = np.empty((3,))
                    for term in range(3):
                        _,_,passed[term] = test_randstats(stat, method, test=test, term=term,
                                                          do_tests=do_tests, **kwargs)
                    passed = passed.all()

                else:
                    _,_,passed = test_randstats(stat, method, test=test, do_tests=do_tests,
                                                **kwargs)

                print('%s' % 'PASSED' if passed else 'FAILED')
                if 'plot_dir' in kwargs: plt.close('all')


# =============================================================================
# Tests for functions computing confidence intervals
# =============================================================================
def test_confints(stat, test='gain', test_values=None, distribution='normal', confint=0.95,
                  n_reps=100, seed=None, do_tests=True, do_plots=False, plot_dir=None, **kwargs):
    """
    Basic testing for bootstrap confidence interval computation functions

    Generate synthetic data, computes statistics, confints, and significance using given method,
    and compare computed to expected values.

    For test failures, raises an error or warning (depending on value of `do_tests`).
    Optionally plots summary of test results.
    
    Parameters
    ----------
    stat : str
        Type of statistical test to evaluate: 'one_sample' | 'paired_sample' | 'two_sample'

    test : str, default: 'gain'
        Type of test to run. Options:
        
        - 'gain' : Tests multiple values for btwn-cond response difference (gain).
            Checks for monotonically increasing confints.
        - 'spread' : Tests multiple values for distribution spread (SD).
            Checks for monotonically decreasing confints.
        - 'n' : Tests multiple values of number of trials (n).
            Checks that confints decrease with n.
        - 'bias' : Tests multiple n values with 0 btwn-cond difference.
            Checks that stat doesn't vary and p value remains ~ 0 (unbiased).

    test_values : array-like, shape=(n_values,), dtype=str
        List of values to test. Interpretation and defaults are test-specific:
        
        - 'gain' :      Btwn-condition response differences (gains). Default: [1,2,5,10,20]
        - 'spread' :    Gaussian SDs for each response distribution. Default: [1,2,5,10,20]
        - 'n'/'bias' :  Trial numbers. Default: [25,50,100,200,400,800]

    distribution : str, default: 'normal'
        Name of distribution to simulate data from. Options: 'normal' | 'poisson'

    confint : float, default: 0.95 (95% CI)
        Confidence interval to compute (1-alpha). 

    do_tests : bool, default: True
        Set=True to evaluate test results against expected values and raise an error if they fail

    do_plots : bool, default: False
        Set=True to plot test results

    plot_dir : str, default: None (don't save to file)
        Full-path directory to save plots to. Set=None to not save plots.

    seed : int, default: 1 (reproducible random numbers)
        Random generator seed for repeatable results. Set=None for fully random numbers.

    **kwargs :
        All other keyword args passed as-is to `simulate_dataset` or confint computation
        function, as appropriate

    Returns
    -------
    means : dict, {str : ndarray. shape=(n_values,)}
        Mean results (across independent test runs) of variables output from randomization tests
        for each tested value. Each key/value pair corressponds to a computed statistic variable:
        
        - 'signif' :       Binary signficance decision (at criterion <alpha>)
        - 'p' :            p values (actually -log10(p) so increases with effect size)
        - 'stat_obs' :     Observed evaluatation statistic values
        - 'stat_resmp' :   Mean resampled statistic values (across all resamples)

    sds : {str : (n_values,) ndarray}
        Standard deviation of results (across test runs) of variables output from rand tests
        for each tested value. Same fields as means.

    passed : bool
        True if all tests produce expected values; otherwise False.
    """
    # HACK Hard-code term=0, but keep term for possible future extension to 1-way/2-way stats
    term = 0
    # Note: Set random seed once here, not for every random data generation loop below
    if seed is not None: set_random_seed(seed)

    test = test.lower()

    # Set defaults for tested values and set up data generator function depending on <test>
    # Note: Only set random seed once above, don't reset in data generator function calls
    if stat == 'two_way':       n_conds = 4         # Simulate 2-way design w/ 4 conds (2x2)
    elif stat == 'one_sample':  n_conds = 1
    else:                       n_conds = 2
    if stat == 'two_way':
        # Set gains for effect on 1st main effect, 2nd main effect, interaction effect
        if term == 0:   gain_pattern = np.asarray([0,0,1,1])
        elif term == 1: gain_pattern = np.asarray([1,0,1,0])
        elif term == 2: gain_pattern = np.asarray([0,1,1,0])
    else:
        gain_pattern = 1
    sim_args = dict(gain=5.0*gain_pattern, offset=0.0, spreads=10.0, n_conds=n_conds, n=100,
                    distribution=distribution, seed=None)
    # Override defaults with any simulation-related params passed to function
    for arg in sim_args:
        if arg in kwargs: sim_args[arg] = kwargs.pop(arg)

    if test == 'gain':
        test_values = [1,2,5,10,20] if test_values is None else test_values
        del sim_args['gain']                        # Delete preset arg so uses arg to lambda below
        gen_data = lambda gain: simulate_dataset(**sim_args,gain=gain*gain_pattern)

    elif test in ['spread','spreads','sd']:
        test_values = [1,2,5,10,20] if test_values is None else test_values
        del sim_args['spreads']                     # Delete preset arg so uses arg to lambda below
        gen_data = lambda spreads: simulate_dataset(**sim_args,spreads=spreads)

    elif test in ['n','n_trials','bias']:
        test_values = [25,50,100,200,400,800] if test_values is None else test_values
        if test == 'bias': sim_args.update(offset=0, gain=0)    # Set gain,offset=0 for bias test
        del sim_args['n']                           # Delete preset arg so uses arg to lambda below
        gen_data = lambda n_trials: simulate_dataset(**sim_args,n=n_trials)

    elif test == 'n_conds':
        test_values = [2,4,8] if test_values is None else test_values
        del sim_args['n_conds']                     # Delete preset arg so uses arg to lambda below
        gen_data = lambda n_conds: simulate_dataset(**sim_args,n_conds=n_conds)

    else:
        raise ValueError("Unsupported value '%s' set for <test>" % test)

    # Set up function for computing randomization stats
    confint_func = _str_to_confint_func(stat)
    kwargs.update(return_stats=True)
    kwargs.update(confint=confint)
    if 'n_resamples' not in kwargs: kwargs['n_resamples'] = 100

    n_terms = 3 if stat == 'two_way' else 1
    results = dict(confints = np.empty((n_terms,2,len(test_values),n_reps)),
                   ci_diff = np.empty((n_terms,len(test_values),n_reps)),
                   signif = np.empty((n_terms,len(test_values),n_reps),dtype=bool),
                   stat_obs = np.empty((n_terms,len(test_values),n_reps)),
                   stat_resmp = np.empty((n_terms,len(test_values),n_reps)))

    resmp_axis = -1 if stat == 'two_way' else 0

    for i_value,test_value in enumerate(test_values):
        for i_rep in range(n_reps):
            # Generate simulated data with current test value
            data,labels = gen_data(test_value)

            # One-sample tests
            if stat == 'one_sample':
                confints, stat_obs, stat_resmp = confint_func(data, **kwargs)

            # 1-way and 2-way ANOVA-like multi-level factorial tests
            elif stat in ['one_way','two_way']:
                # For 2-way tests, reorg labels from (n,) vector of values in set {0-3}
                # to (n,3) array where 1st 2 col's are 2 orthogonal factors, 3rd col is interaction
                if stat == 'two_way':
                    labels = np.stack((labels >= 2, (labels == 1) | (labels == 3), labels), axis=1)
                confints, stat_obs, stat_resmp = confint_func(data, labels, **kwargs)

            # Paired-sample and two-sample tests
            else:
                confints, stat_obs, stat_resmp = confint_func(data[labels==1], data[labels==0],
                                                              **kwargs)

            # Determine which values are significant (p < alpha criterion)
            results['signif'][:,i_value,i_rep] = (confints > 0).all()
            # Negative log-transform p values so increasing = "better" and less compressed
            results['confints'][:,:,i_value,i_rep] = confints
            results['stat_obs'][:,i_value,i_rep] = stat_obs
            # Compute mean resampled stat value across all resamples
            results['stat_resmp'][:,i_value,i_rep] = stat_resmp.mean(axis=resmp_axis)

    # Compute difference btwn each set of confidence intervals (upper - lower)
    results['ci_diff'] = np.diff(results['confints'], axis=1).squeeze(axis=1)

    # Compute mean and std dev across different reps of simulation
    means   = {variable : values.mean(axis=-1) for variable,values in results.items()}
    sds     = {variable : values.std(axis=-1,ddof=0) for variable,values in results.items()}
    variables = list(means.keys())

    if do_plots:
        plt.figure()
        for i_vbl,variable in enumerate(variables):
            for i_term in range(n_terms):
                mean = means[variable][i_term,:]
                sd = sds[variable][i_term,:]

                sp = i_term*len(variables) + i_vbl + 1
                plt.subplot(n_terms, len(variables), sp)
                plt.grid(axis='both',color=[0.75,0.75,0.75],linestyle=':')
                if variable == 'confints':
                    # Plot lower, upper confints separately
                    plt.errorbar(test_values, mean[0,:], sd[0,:], marker='o')
                    plt.errorbar(test_values, mean[1,:], sd[1,:], marker='o')
                else:
                    plt.errorbar(test_values, mean, sd, marker='o')
                if i_term == n_terms-1:
                    plt.xlabel('n' if test == 'bias' else test)
                    plt.ylabel(variable)
                if i_term == 0: plt.title(variable)
                if stat == 'two_way': plt.ylabel("Term %d" % i_term)

        if plot_dir is not None:
            filename = os.path.join(plot_dir,'stat-summary-%s-%s' % (stat,test))
            if stat == 'two_way': filename += '-term%d' % term
            plt.savefig(filename)

    # Determine if test actually produced the expected values
    # 'gain' : Test if statistic and significance increases monotonically with between-group gain
    if test == 'gain':
        evals = [((np.diff(means['signif'][term,:]) >= 0).all(),
                    "Significance does not increase monotonically with btwn-cond mean diff"),
                 (means['ci_diff'][term,:].ptp() < sds['ci_diff'][term,:].max(),
                    "Confints have larger than expected range with btwn-cond mean diff"),
                 ((np.diff(means['stat_obs'][term,:]) > 0).all(),
                    "Statistic does not increase monotonically with btwn-cond mean diff"),
                 ((np.diff(means['stat_resmp'][term,:]) > 0).all(),
                    "Resampled statistic does not increase monotonically with btwn-cond mean diff")]
        # Confidence intervals not expected to change with mean for 1-sample data, so remove test
        if stat == 'one_sample': evals.pop(1)

    # 'spread' : Test if confints increase and stat decreases monotonic with within-group spread
    elif test in ['spread','spreads','sd']:
        evals = [((np.diff(means['signif'][term,:]) > 0).all(),
                    "Signif does not decrease monotonically with within-cond spread increase"),
                 ((np.diff(means['ci_diff'][term,:]) > 0).all(),
                    "Confints do not increase monotonically with within-cond spread increase"),
                 ((np.diff(means['stat_obs'][term,:]) < 0).all(),
                    "Statistic does not decrease monotonically with within-cond spread increase"),
                 ((np.diff(means['stat_resmp'][term,:]) < 0).all(),
                    "Resampled stat does not decrease monotonic with within-cond spread increase")]

    # 'n' : Test if confints decrease, but statistic is ~ same for all values of n (unbiased by n)
    elif test in ['n','n_trials']:
        evals = [((np.diff(means['signif'][term,:]) >= 0).all(),
                    "Significance does not increase monotonically with n"),
                 ((np.diff(means['ci_diff'][term,:]) <= 0).all(),
                    "Confints do not decrease monotonically with n"),
                 (means['stat_obs'][term,:].ptp() <= sds['stat_obs'][term,:].max(),
                    "Statistic has larger than expected range with n"),
                 (means['stat_resmp'][term,:].ptp() <= sds['stat_resmp'][term,:].max(),
                    "Resampled statistic has larger than expected range with n")]

    # 'bias': Test that statistic is not > 0 and p value ~ alpha if gain = 0, for varying n
    elif test == 'bias':
        evals = [((means['confints'][term,0,:] < 0).all() & (0 < means['confints'][term,1,:]).all(),
                    "Confints don't overlap with 0 when no mean diff between conditions"),
                 ((np.abs(means['stat_obs'][term,:]) < sds['stat_obs'][term,:].max()).all(),
                    "Statistic is above 0 when no mean diff between conditions"),
                 ((np.abs(means['stat_resmp'][term,:]) < sds['stat_resmp'][term,:].max()).all(),
                    "Resampled statistic is above 0 when no mean diff between conditions")]

    passed = True
    for cond,message in evals:
        if not cond:    passed = False

        # Raise an error for test fails if do_tests is True
        if do_tests:    assert cond, AssertionError(message)
        # Just issue a warning for test fails if do_tests is False
        elif not cond:  print("Warning: " + message)

    return means, sds, passed


def confint_test_battery(stats=['one_sample','paired_sample','two_sample'],
                         tests=('gain','n','bias'), do_tests=True, **kwargs):
    """
    Run a battery of given tests on given bootstrap confidence interval computation methods

    Parameters
    ----------
    stats : array-like of str, default: ['one_sample','paired_sample','two_sample']
        List of statistical tests to evaluate.
                

    tests : array-like of str, default: ('gain','n','bias') (all supported tests)
        List of tests to run.
                
    tests : array-like of str, default: ('gain','spread','n','bias','correlation')
        List of tests to run.
                
    do_tests : bool, default: True
        Set=True to evaluate test results against expected values and raise an error if they fail

    **kwargs :
        Any other kwargs passed directly to test_confints()
    """
    if isinstance(stats,str): stats = [stats]
    if isinstance(tests,str): tests = [tests]

    for stat in stats:
        for test in tests:
            print("Running %s test on %s" % (test,stat))

            # Run separate tests for each term of 2-way stats (2 main effects and interaction)
            if stat == 'two_way':
                passed = np.empty((3,))
                for term in range(3):
                    _,_,passed[term] = test_confints(stat, test=test, term=term,
                                                     do_tests=do_tests, **kwargs)
                passed = passed.all()

            else:
                _,_,passed = test_confints(stat, test=test, do_tests=do_tests, **kwargs)

            print('%s' % 'PASSED' if passed else 'FAILED')
            if 'plot_dir' in kwargs: plt.close('all')


# =============================================================================
# Helper functions
# =============================================================================
def _str_to_stat_func(stat):
    """ Convert string specifier for statistic to function for computing it """
    stat = stat.lower()
    if stat == 'one_sample':            return one_sample_test
    elif stat == 'paired_sample':       return paired_sample_test
    elif stat == 'paired_sample_assoc': return paired_sample_association_test
    elif stat == 'two_sample':          return two_sample_test
    elif stat == 'one_way':             return one_way_test
    elif stat == 'two_way':             return two_way_test
    else:
        raise ValueError("Unknown stat type '%s'" % stat)


def _str_to_confint_func(stat):
    """ Convert string specifier for statistic to function for computing its confints """
    stat = stat.lower()
    if stat == 'one_sample':        return one_sample_confints
    elif stat == 'paired_sample':   return paired_sample_confints
    elif stat == 'two_sample':      return two_sample_confints
    else:
        raise ValueError("Unknown stat type '%s'" % stat)



# -*- coding: utf-8 -*-
""" Continuous-wavelet-based spectral analysis """
from math import sqrt, pi, ceil
import numpy as np

from spynal.utils import standardize_array
from spynal.spikes import _spike_data_type, times_to_bool
from spynal.spectra.preprocess import remove_dc
from spynal.spectra.utils import next_power_of_2, complex_to_spec_type
from spynal.spectra.helpers import fft, ifft, _FFTW_KWARGS_DEFAULT, \
                                   _extract_triggered_data, _undo_standardize_array_newaxis


def wavelet_spectrum(data, smp_rate, axis=0, data_type='lfp', spec_type='complex', freqs=None,
                     removeDC=True, wavelet='morlet', wavenumber=6, pad=False, buffer=0, **kwargs):
    """
    Compute continuous wavelet transform of data, then averages across timepoints to
    reduce it down to a frequency spectrum.

    Not really the best way to compute 1D frequency spectra, but included for completeness

    Only parameters differing from :func:`.spectrum` are described here.

    Parameters
    ----------
    freqs : array-like, shape=(n_freqs,), default: 2**np.arange(1,7.5,0.25)
        Set of desired wavelet frequencies. Default value logarithmically samples from 2-152
        in 1/4 octaves, but log sampling is not required.

    wavelet : {'morlet'}, default: 'morlet'
        Name of wavelet type. Currently only 'morlet' is supported.

    wavenumber : int, default: 6
        Wavelet wave number parameter ~ number of oscillations in each wavelet.
        Must be >= 6 to meet "admissibility constraint".

    buffer : float, default: 0 (no buffer)
        Time (s) to trim off each end of time dimension of data.
        Removes symmetric buffer previously added (outside of here) to prevent edge effects.

    Returns
    -------
    spec : ndarray, shape=(...,n_freqs,...), dtype=complex or float.
        Wavelet-derived spectrum of data.
        Same shape as data, with frequency axis replacing time axis
        dtype is complex if `spec_type` is 'complex', float otherwise.

    freqs, ndarray, shape=(n_freqs,)
        List of frequencies in `spec` (in Hz)
    """
    if freqs is None: freqs = 2.0**np.arange(1,7.5,0.25)
    if axis < 0: axis = data.ndim + axis

    spec, freqs, _ = wavelet_spectrogram(data, smp_rate, axis=axis, data_type=data_type,
                                         spec_type=spec_type, freqs=freqs, removeDC=removeDC,
                                         wavelet=wavelet, wavenumber=wavenumber, pad=pad,
                                         buffer=buffer, **kwargs)

    # Take mean across time axis (which is now shifted +1 b/c of frequency axis)
    return spec.mean(axis=axis+1), freqs


def wavelet_spectrogram(data, smp_rate, axis=0, data_type='lfp', spec_type='complex', freqs=None,
                        removeDC=True, wavelet='morlet', wavenumber=6, pad=False, buffer=0,
                        downsmp=1, **kwargs):
    """
    Compute continuous time-frequency wavelet transform of data at given frequencies.

    Only parameters differing from :func:`.spectrogram` are described here.

    Parameters
    ----------
    freqs : array-like, shape=(n_freqs,), default: 2**np.arange(1,7.5,0.25)
        Set of desired wavelet frequencies. Default value logarithmically samples from 2-152
        in 1/4 octaves, but log sampling is not required.

    wavelet : {'morlet'}, default: 'morlet'
        Name of wavelet type. Currently only 'morlet' is supported.

    wavenumber : int, default: 6
        Wavelet wave number parameter ~ number of oscillations in each wavelet.
        Must be >= 6 to meet "admissibility constraint".

    buffer : float, default: 0 (no buffer)
        Time (s) to trim off each end of time dimension of data.
        Removes symmetric buffer previously added (outside of here) to prevent edge effects.

    downsmp: int, default: 1 (no downsampling)
        Factor to downsample time sampling by (after spectral analysis).
        eg, smp_rate=1000 (dt=0.001), downsmp=10 -> smpRateOut=100 (dt=0.01)

    Returns
    -------
    spec : ndarray, shape=(...,n_freqs,n_timepts_out,...), dtype=complex or float.
        Wavelet time-frequency spectrogram of data, transformed to requested spectral type.
        Same shape as data, with frequency axis prepended before time, and time axis
        possibly reduces via downsampling.
        dtype is complex if `spec_type` is 'complex', float otherwise.

    freqs : ndarray, shape=(n_freqs,) ndarray
        List of frequencies in `spec` (in Hz)

    timepts : ndarray, shape=(n_timepts_out,)
        List of timepoints in `spec` (in s, referenced to start of data).

    References
    ----------
    Torrence & Compo 1998 https://doi.org/10.1175/1520-0477(1998)079%3C0061:APGTWA%3E2.0.CO;2
    """
    if freqs is None: freqs = 2.0**np.arange(1,7.5,0.25)
    if axis < 0: axis = data.ndim + axis

    # Convert spike timestamp data to boolean spike train format
    if (data_type == 'spike') and (_spike_data_type(data) == 'timestamp'):
        lims    = kwargs.pop('lims',None)
        bins    = kwargs.pop('bins',None)
        data,_  = times_to_bool(data, width=1/smp_rate, lims=lims, bins=bins)
        axis    = data.ndim
    assert len(kwargs) == 0, \
        TypeError("Incorrect or misspelled variable(s) in keyword args: "+', '.join(kwargs.keys()))

    # Convert buffer from s -> samples
    if buffer != 0:  buffer  = int(ceil(buffer*smp_rate))

    # Reshape data array -> (n_timepts_in,n_dataseries) matrix
    data, data_shape = standardize_array(data, axis=axis, target_axis=0)
    n_timepts_in = data.shape[0]

    # Time indexes to extract from spectrogram for output (accounting for buffer, downsampling)
    time_idxs_out = np.arange(buffer,n_timepts_in-buffer,downsmp)

    # Set FFT length = data length if no padding; else pad to next power of two
    if not pad: n_fft = n_timepts_in
    else:       n_fft = next_power_of_2(n_timepts_in)

    # Compute set of Fourier-transformed wavelet functions (if not already given)
    if isinstance(wavelet,str):
        wavelets_fft = compute_wavelets(n_fft,smp_rate,freqs=freqs,
                                        wavelet=wavelet,wavenumber=wavenumber,
                                        do_fft=True)
    else:
        wavelets_fft = wavelet

    if removeDC: data = remove_dc(data,axis=0)

    # Compute FFT of data
    data = fft(data, n=n_fft,axis=0, **_FFTW_KWARGS_DEFAULT)

    # Reshape data -> (1,n_timepts,n_series) (insert axis 0 for wavelet scales/frequencies)
    # Reshape wavelets -> (n_freqs,n_timepts,1) to broadcast
    #  (except for special case of 1D data with only a single time series)
    data = data[np.newaxis,...]
    if data.ndim == 3: wavelets_fft = wavelets_fft[:,:,np.newaxis]

    # Convolve data with wavelets (multiply in Fourier domain)
    # -> inverse FFT to get wavelet transform
    spec = ifft(data*wavelets_fft, n=n_fft,axis=1, **_FFTW_KWARGS_DEFAULT)[:,time_idxs_out,...]

    # Convert to desired output spectral signal type
    spec    = complex_to_spec_type(spec,spec_type)

    spec    = _undo_standardize_array_newaxis(spec,data_shape,axis=axis)

    timepts = time_idxs_out.astype(float)/smp_rate  # Convert time sampling from samples -> s

    return spec, freqs, timepts


def compute_wavelets(n, smp_rate, freqs=None, wavelet='morlet', wavenumber=6, do_fft=False):
    """
    Compute set of (Fourier transformed) wavelets for use in wavelet spectral analysis

    Parameters
    ----------
    n : int
        Total number of samples (time points) in analysis, including any padding.

    smp_rate scalar
        Data sampling rate (Hz)

    freqs : array-like, shape=(n_freqs,), default: 2**np.arange(1,7.5,0.25)
        Set of desired wavelet frequencies. Default value logarithmically samples from 2-152
        in 1/4 octaves, but log sampling is not required.

    wavelet : {'morlet'}, default: 'morlet'
        Name of wavelet type. Currently only 'morlet' is supported.

    wavenumber : int, default: 6
        Wavelet wave number parameter ~ number of oscillations in each wavelet.
        Must be >= 6 to meet "admissibility constraint".

    do_fft: bool, default: False
        If True, returns Fourier transform of wavelet functions.
        If False, returns original time-domain functions.

    Returns
    -------
    wavelets : ndarray, shape=(n_freqs,n_timepts)
        Computed set of wavelet functions at multiple frequencies/scales.
        (either the time domain wavelets or theirFourier transform, depending on `do_fft`)

    References
    ----------
    Torrence & Compo 1998 https://doi.org/10.1175/1520-0477(1998)079%3C0061:APGTWA%3E2.0.CO;2
    """
    if freqs is None:   freqs = 2.0**np.arange(1,7.5,0.25)
    else:               freqs = np.asarray(freqs)
    wavelet = wavelet.lower()
    dt      = 1/smp_rate         # Convert sampling rate -> sampling interval

    if wavelet == 'morlet':
        # Conversion factor from scale to Fourier period for Morlet wavelets [T&C Table 1]
        # period = 1/frequency = scale * scale_to_period
        scale_to_period = (4.0*pi)/(wavenumber + sqrt(2.0 + wavenumber**2))

        # Convert desired frequencies -> scales for full set of wavelets
        scales          = (1.0/freqs) / scale_to_period
        scales          = scales[:,np.newaxis] # -> (n_freqs,1)

        # Construct wavenumber array used in transform [T&C Eqn(5)]
        k   = np.arange(1, int(np.fix(n/2.0)+1))
        k   = k*((2.0*pi)/(n*dt))
        k   = np.hstack((np.zeros((1,)), k, -k[int(np.fix((n-1)/2)-1) : : -1]))
        k   = k[np.newaxis,:]   # -> (1,n_timepts_in)
        k0  = wavenumber

        # Total energy=N   [T&C Eqn(7)]
        normalization   = np.sqrt(scales*k[:,1])*(pi**(-0.25))*sqrt(n)
        # Wavelet exponent
        exponent        = -0.5*(scales*k - k0)**2 * (k > 0)

        # Fourier transform of Wavelet function
        if do_fft:
            wavelets = normalization*np.exp(exponent) * (k > 0)
        else:
            raise NotImplementedError("non-FFT wavelet output not coded up yet (TODO)")

    else:
        raise ValueError("Unsupported value '%s' given for <wavelet>." \
                         "Currently only 'Morlet' suppported")

    return wavelets


def wavelet_bandwidth(freqs, wavelet='morlet', wavenumber=6, full=True):
    """
    Return frequency and time bandwidths for set of wavelets at given frequencies

    Parameters
    ----------
    freqs : array-like, shape=(n_freqs,)
        Set of wavelet center frequencies.

    wavelet : {'morlet'}, default: 'morlet'
        Name of wavelet type. Currently only 'morlet' is supported.

    wavenumber : int, default: 6
        Wavelet wave number parameter ~ number of oscillations in each wavelet.

    full : bool, default: True
        If True, return full-bandwidths. If False, return half-bandwidths.

    Returns
    -------
    freq_widths : ndarray, shape=(n_freqs,)
        Frequency bandwidth (Hz) for each given frequency

    time_widths : ndarray, shape=(n_freqs,)
        Time bandwidth (s) for each given frequency
    """
    wavelet = wavelet.lower()
    freqs = np.asarray(freqs)

    if wavelet == 'morlet':
        freq_widths = freqs / wavenumber
        time_widths = 1 / (2*pi*freq_widths)

    else:
        raise ValueError("Unsupported value '%s' given for <wavelet>." \
                         "Currently only 'Morlet' suppported")

    # Convert half-bandwidths to full-bandwidths
    if full:
        freq_widths = 2 * freq_widths
        time_widths = 2 * time_widths

    return freq_widths, time_widths


def wavelet_edge_extent(freqs, wavelet='morlet', wavenumber=6):
    """
    Return temporal extent of edge effects for set of wavelets at given frequencies

    Compute time period over which edge effects might effect output of wavelet transform,
    and over which the effects of a single spike-like artifact in data will extend.

    Computed as time for wavelet power to drop by a factor of exp(−2), ensuring that
    edge effects are "negligible" beyond this point.

    Parameters
    ----------
    freqs : array-like, shape=(n_freqs,)
        Set of wavelet center frequencies

    wavelet : {'morlet'}, default: 'morlet'
        Name of wavelet type. Currently only 'morlet' is supported.

    wavenumber : int, default: 6
        Wavelet wave number parameter ~ number of oscillations in each wavelet.

    Returns
    -------
    edge_extent : ndarray, shape=(n_freqs,)
        Time period (s) over which edge effects extend for each given frequency

    References
    ----------
    Torrence & Compo https://doi.org/10.1175/1520-0477(1998)079%3C0061:APGTWA%3E2.0.CO;2 Sxn.3g
    """
    wavelet = wavelet.lower()
    freqs = np.asarray(freqs)

    if wavelet == 'morlet':
        # "Fourier factor" (conversion factor from scale to Fourier period [T&C Table 1])
        scale_to_period = (4.0*pi)/(wavenumber + sqrt(2.0 + wavenumber**2))
        # Convert given frequencies -> scales for set of wavelets
        scales          = (1.0/freqs) / scale_to_period
        # Cone-of-influence for Morlet = sqrt(2)*scale [T&C Table 1]
        edge_extent     = sqrt(2.0) * scales

    else:
        raise ValueError("Unsupported value '%s' given for <wavelet>." \
                         "Currently only 'Morlet' suppported")

    return edge_extent




# -*- coding: utf-8 -*-
""" Helper functions for `sync` module for oscillatory synchrony analysis """
import numpy as np

from spynal.utils import index_axis
from spynal.spectra.spectra import spectrogram
 

def _infer_data_type(data):
    """ Infer type of data signal given -- 'raw' (real) | 'spectral' (complex) """
    if np.isrealobj(data):  return 'raw'
    else:                   return 'spectral'


def _sync_raw_to_spectral(data1, data2, smp_rate, axis, time_axis, taper_axis,
                          spec_method, data_type, **kwargs):
    """
    Check data input to lfp-lfp synchrony methods.
    Determine if input data is raw or spectral, compute spectral transform if raw.
    """
    if data_type is None: data_type = _infer_data_type(data1)

    assert _infer_data_type(data2) == data_type, \
        ValueError("data1 and data2 must have same data_type (raw or spectral)")

    # If raw data is input, compute spectral transform first
    if data_type == 'raw':
        assert smp_rate is not None, \
            "For raw/time-series data, need to input value for <smp_rate>"
        assert time_axis is not None, \
            "For raw/time-series data, need to input value for <time_axis>"
        if spec_method == 'multitaper': kwargs.update(keep_tapers=True)
        data1,freqs,timepts = spectrogram(data1, smp_rate, axis=time_axis, method=spec_method,
                                          data_type='lfp', spec_type='complex', **kwargs)
        data2,freqs,timepts = spectrogram(data2, smp_rate, axis=time_axis, method=spec_method,
                                          data_type='lfp', spec_type='complex', **kwargs)
        # Account for new frequency (and/or taper) axis prepended before time_axis
        n_new_axes = 2 if spec_method == 'multitaper' else 1
        if axis >= time_axis: axis += n_new_axes
        time_axis += n_new_axes
        if spec_method == 'multitaper': taper_axis = time_axis-1

    else:
        freqs = []
        timepts = []
        if spec_method == 'multitaper':
            assert taper_axis is not None, \
                ValueError("Must set value for taper_axis for multitaper spectral inputs")

    return data1, data2, freqs, timepts, axis, time_axis, taper_axis


def _sfc_raw_to_spectral(spkdata, lfpdata, smp_rate, axis, time_axis, taper_axis, timepts,
                         method, spec_method, data_type, **kwargs):
    """
    Check data input to spike-field coupling methods.
    Determine if input data is raw or spectral, compute spectral transform if raw.
    """
    if data_type is None:
        spk_data_type = _infer_data_type(spkdata)
        lfp_data_type = _infer_data_type(lfpdata)
        # Spike and field data required to have same type (both raw or spectral) for coherence
        if method == 'coherence':
            assert spk_data_type == lfp_data_type, \
                ValueError("Spiking (%s) and LFP (%s) data must have same data type" % \
                            (spk_data_type,lfp_data_type))
        # Spike data must be raw (not spectral) for phase-based methods
        else:
            assert _infer_data_type(spkdata) == 'raw', \
                ValueError("Spiking data must be given as raw, not spectral, data")

        data_type = lfp_data_type

    # If raw data is input, compute spectral transform first
    if data_type == 'raw':
        assert (timepts is not None) or (smp_rate is not None), \
            ValueError("If no value is input for <timepts>, must input value for <smp_rate>")

        # Ensure spkdata is boolean array with 1's for spiking times (ie not timestamps)
        assert spkdata.dtype != object, \
            TypeError("Spiking data must be converted from timestamps to boolean format")
        spkdata = spkdata.astype(bool)

        # Default timepts to range from 0 - n_timepts/smp_rate
        if timepts is None:     timepts = np.arange(lfpdata.shape[time_axis]) / smp_rate
        elif smp_rate is None:  smp_rate = 1 / np.diff(timepts).mean()

        # For multitaper, keep tapers, to be averaged across like trials below
        if spec_method == 'multitaper': kwargs.update(keep_tapers=True)
        # For multitaper phase sync, spectrogram window spacing must = sampling interval (eg 1 ms)
        if (spec_method == 'multitaper') and (method != 'coherence'):
            kwargs.update(spacing=1/smp_rate)

        # All spike-field coupling methods require spectral data for LFPs
        lfpdata,freqs,times = spectrogram(lfpdata, smp_rate, axis=time_axis, method=spec_method,
                                          data_type='lfp', **kwargs)

        # Coherence requires spectral data for both spikes and LFPs
        if method == 'coherence':
            spkdata,_,_ = spectrogram(spkdata, smp_rate, axis=time_axis, method=spec_method,
                                      data_type='spike', **kwargs)

        timepts_raw = timepts
        timepts = times + timepts[0]

        # Multitaper spectrogram loses window width/2 timepoints at either end of data
        # due to windowing. For phase-based SFC methods, must remove these timepoints
        # from spkdata to match. (Note: not issue for coherence, which operates in spectral domain)
        if (spec_method == 'multitaper') and (method != 'coherence'):
            retained_times = (timepts_raw >= timepts[0]) & (timepts_raw <= timepts[-1])
            spkdata = index_axis(spkdata, time_axis, retained_times)

        # Frequency axis always inserted just before time axis, so if
        # observation/trial axis is later, must increment it
        # Account for new frequency (and/or taper) axis
        n_new_axes = 2 if spec_method == 'multitaper' else 1

        if method != 'coherence':
            # Set up indexing to preserve axes before/after time axis,
            # but insert n_new_axis just before it
            slicer = [slice(None)]*time_axis + \
                    [np.newaxis]*n_new_axes + \
                    [slice(None)]*(spkdata.ndim-time_axis)
            # Insert singleton dimension(s) into spkdata to match freq/taper dim(s) in lfpdata
            spkdata = spkdata[tuple(slicer)]

        if axis >= time_axis: axis += n_new_axes
        time_axis += n_new_axes
        if spec_method == 'multitaper': taper_axis = time_axis - 1

    else:
        freqs = []
        # timepts = []

    return spkdata, lfpdata, freqs, timepts, smp_rate, axis, time_axis, taper_axis
